extern const char* X_GAME_PROFILE_xpm[];
extern const char* M_WND_ICON128_xpm[];
extern const char* X_BOX_xpm[];
extern const char* X_SETTINGS_xpm[];

extern unsigned char PNG_CHECK_YES_png[573];
extern unsigned char PNG_ERROR_png[472];
extern unsigned char PNG_HELP_png[492];
extern unsigned char PNG_REFRESH_png[449];

extern unsigned char DEBUGGER_BP_RED_png[470];
extern unsigned char DEBUGGER_PAUSE_png[211];
extern unsigned char DEBUGGER_PLAY_png[306];
extern unsigned char DEBUGGER_STEP_INTO_png[297];
extern unsigned char DEBUGGER_STEP_OUT_png[308];
extern unsigned char DEBUGGER_STEP_OVER_png[296];
extern unsigned char DEBUGGER_BP_png[436];
extern unsigned char DEBUGGER_GOTO_png[292];

extern unsigned char INPUT_CONNECTED_png[400];
extern unsigned char INPUT_DISCONNECTED_png[410];
extern unsigned char INPUT_LOW_BATTERY_png[340];