// Fill out your copyright notice in the Description page of Project Settings.

#include "unrealstatus.h"
#include "Modules/ModuleManager.h"

IMPLEMENT_PRIMARY_GAME_MODULE(FDefaultGameModuleImpl, unrealstatus, "unrealstatus");
