// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/GameModeBase.h"
#include "unrealstatusGameModeBase.generated.h"

/**
 *
 */
UCLASS()
class UNREALSTATUS_API AunrealstatusGameModeBase : public AGameModeBase {
    GENERATED_BODY()
};
